/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : SetCoverProblem.cc                                       */
/*                                                                       */
/*       Date : March  4, 1998                                           */
/* Updated at : March 26, 1998                                           */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <iomanip.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <deffunc.h>

#include "SetCoverProblem.h"

/*************************************************************************/
/*************************************************************************/

SetCoverProblem::SetCoverProblem(ulong P1,ulong P2,ulong P3,ulong P4,double P5)
{   int i; 
    NumOfRows = P1 ;
    NumOfColumns =P2;
    NumberOfNonZeros =P3;
    TheBestIP =P4;
    TheBestLP =P5;

}// Constructor

/*************************************************************************/
/*************************************************************************/

SetCoverProblem::SetCoverProblem(const SetCoverProblem& RHS)
{    
    NumOfRows = RHS.NumOfRows ;
    NumOfColumns = RHS.NumOfColumns;
    NumberOfNonZeros = RHS.NumberOfNonZeros ;
    TheBestIP = RHS.TheBestIP;
    TheBestLP = RHS.TheBestLP;
    myColumns = RHS.myColumns;

}// COPY Constructor

/*************************************************************************/
/** Adds a new column to the Problem Set.  */
/*************************************************************************/

void 
SetCoverProblem::UpdateRow( ulong ColumnNumber, ulong NumOfNZeros, 
                            ulong _profit, ulong col_rep, int *array)
{  
   Column *p1;
   int i;
   
   if( ColumnNumber>=0 && ColumnNumber<NumOfColumns)  
   {  
      p1 = new Column(NumOfRows);
      p1->setProfit(_profit);
      p1->setIndex(ColumnNumber);

      for(i=0;i<NumOfNZeros;i++) p1->set(array[i]); 

      myColumns.addElement(p1);

      //cout << *myColumns[ColumnNumber]  << endl;

   }// end of if

}// end of UpdateRow() 

/*************************************************************************/
/**
 Finds out the all the minimum cost columns which covers the
 corresponding row.
*/
/*************************************************************************/

void SetCoverProblem::findMinimums() 
{  
   int i,j;
   smallest = new unsigned int[NumOfRows]; 
   cost = new cost_t[NumOfRows]; 
   Column *f;

   for(i=0;i<NumOfRows;i++) smallest[i] = -1;

   for(i=0;i<NumOfColumns;i++)
   {
      if(myColumns.getElementAtSafe(i,f))
      {
	 for(j=0;j<NumOfRows;j++)
	   if((f)->test(j) && (smallest[j]==-1 || cost[j]>(f)->getProfit()) )
	   {
	      smallest[j] = i;
	      cost[j] = (f)->getProfit();
	   
	   }// end of if

      }// end of if

   }// end of for

}// end of findMinimums()

/*************************************************************************/
/** Returns the minimum cost for covering given row.  */
/*************************************************************************/

cost_t SetCoverProblem::getMinimumCostForRowIndex(unsigned int indx) const
{
    cost_t ret = (cost_t) 0;

    if(indx>=0 && indx<NumOfRows && smallest[indx]>-1) ret = cost[indx];

    return ret;

}// end of getMinimumCostForRowIndex()

/*************************************************************************/
/** Prints all the available Columns in the problem.  */
/*************************************************************************/

void SetCoverProblem::printRowInfo(ostream& os) const
{  
   ulong i;
   Column *c;

   for(i=0;i<NumOfColumns;i++) 
       if(myColumns.getElementAtSafe(i,c)) os << i << " : " << *c << endl;

}// end of printRowInfo()

/*************************************************************************/
/** Overloading output operator */
/*************************************************************************/

ostream& operator<<(ostream& os, const SetCoverProblem& RHS)
{  ulong i;

   RHS.printRowInfo(os);

   return os;

}// end of operator<<()

/*************************************************************************/
/** Returns the number of rows.  */
/*************************************************************************/

unsigned int SetCoverProblem::getNumOfRows() 
{ return (unsigned int)NumOfRows; }

/*************************************************************************/
/** Returns the number of available columns.  */
/*************************************************************************/

unsigned int SetCoverProblem::getNumOfCols()
{ return (unsigned int)NumOfColumns; }

/*************************************************************************/
/** Returns the column belanging to given index.  */
/*************************************************************************/

Column* SetCoverProblem::getColumnWithIndx(int idx)
{
   Column *c;

   return ( myColumns.getElementAtSafe(idx,c)?c:NULL);

}// end of getColumnWithIndx

/*************************************************************************/
/*************************************************************************/

SetCoverProblem::~SetCoverProblem()
{
   delete[] smallest;
   delete[] cost;
   Column *c;

   for(int i=0;i<NumOfColumns;i++) if(myColumns.getElementAtSafe(i,c)) delete c;

   if( LookUpTable!=NULL ) delete LookUpTable;

   cout << "deallocate" << endl;

}// end of destructor

/*************************************************************************/
/*************************************************************************/

void SetCoverProblem::sort()
{   
    /*
    ColumnPair *temp; 
    unsigned i,j;

    for(i=0;i<NumOfColumns;i++)
       for(j=i+1;j<NumOfColumns;j++)
	 if(myColumns[i]->getProfit()>myColumns[j]->getProfit())
	 {  temp = myColumns[j];
	    myColumns[j] = myColumns[i];
	    myColumns[i] = temp;
	 }
    */

}// end of sort

/*************************************************************************/
/** Returns -1 if the pattern does not exist.  Exact matching */
/*************************************************************************/

SetCoverProblem::getIndexOfSignature(char *pattern)
{  
   return getIndexOfParSignature(-1,pattern,0);

}// end of getIndexOfSignature()

/*************************************************************************/
/** 
    Returns -1 if the pattern does not exist.  Matching with 
    "do not care" symbols.
    Returns the index of the column which matches with the
    given pattern.
*/
/*************************************************************************/

SetCoverProblem::getIndexOfParSignature(int exCol, char *pattern)
{
   return getIndexOfParSignature(exCol,pattern,rand() % NumOfColumns);

}// enf of getIndexOfParSignature()

/*************************************************************************/
/*************************************************************************/

SetCoverProblem::getIndexOfParSignature(char *pattern)
{
   return getIndexOfParSignature( -1, pattern, rand() % NumOfColumns);

}// enf of getIndexOfParSignature()

/*************************************************************************/
/** Returns -1 if the pattern does not exist.  */
/*************************************************************************/

SetCoverProblem::getIndexOfParSignature(int exCol, char *pattern, int start)
{  
   char *cover = NULL;
   int found = 0;
   unsigned i = 0;
   unsigned k = 0;
   int j, equal;
   int rand_seed = start;
   Column *crr;

   //cout << " I am looking for " << pattern << " except the col ";
   //cout << exCol << endl; 

   i = start % NumOfColumns;

   while( (!found) && k < NumOfColumns )
   {  
      if( i!=exCol && myColumns.getElementAtSafe(i,crr) )
      {
      // get cover of the column

      cover = crr->getCover();

      equal=0;

      if(i!=exCol)
      {  
	 for(j=0;j<NumOfRows;j++) 
	    if(pattern[j]!='#') equal += ( cover[j] == pattern[j] );
	    else equal++;

      }// end of if      

      if( equal == NumOfRows ) found = 1;

      k++;
      i = (i+1) % NumOfColumns;

      /*if(!found) 
      {  cout << setw(4) << i << " ";
	 for(j=0;j<NumOfRows;j++) cout << (cover[j]==1?'1':'0');
	 cout << " does not compatibles with " << endl;
	 cout << setw(4) << exCol << " ";
	 for(j=0;j<NumOfRows;j++) 
	    cout << (pattern[j]!='#'?(pattern[j]==1?'1':'0'):'#');
	 cout << endl;
      }*/

      delete [] cover;
      }// end of if

   }// end of while

   return (found?i:-1);

}// end of getIndexOfParSignature()

/*****************************************************************************/
/** 
    Sorts the column indices set according to their costs in increasing order
*/
/*****************************************************************************/

//tSet<int> SetCoverProblem::sortAccordingToCost(tSet<int>* _set)
void SetCoverProblem::sortAccordingToCost(int tIndx) //tSet<int>* _set)
{
   tSet<int>* _set = LookUpTable+tIndx;
   int setSize = _set->getSize();
   int *indices = new int[setSize];
   tSet<int> ret;
   int val1, val2;
   Column *ci, *cj;
   int i, j, k, tmp;

   //cout << " setSize " << setSize << endl;

   for(i=0;i<setSize;i++) indices[i] = i;

   //cout << *_set << endl;

   for(i=0;i<setSize-1;i++)
   {
      if( _set->getElementAtSafe(indices[i],val1) )
      {
        if( myColumns.getElementAtSafe(val1,ci) )
        {

          for(j=i+1;j<setSize;j++)
          {

	    if( _set->getElementAtSafe(indices[j],val2) )
            {

	      if( myColumns.getElementAtSafe(val2,cj) )
              {
		//cout << " ci: " << *ci << endl;
	        if(cj->getProfit()<ci->getProfit())
	        {
	           //swap(indices[i],indices[j]);
		   tmp = indices[j];
		   indices[j] = indices[i];
		   indices[i] = tmp;
	           ci = cj;

	        }// end of if
	      }// end of if
	    }// end of if

          }// end of for
        }// end of if
      }// end of if
       
   }// end of for

   for(i=0;i<setSize;i++) 
       if(_set->getElementAtSafe(indices[i],val1) ) ret.addElement(val1);

   *_set = ret;

   delete[] indices;

   //return ret;

}// end of sortAccordingToCost()

/*****************************************************************************/
/*****************************************************************************/

void SetCoverProblem::buildLookUpTable()
{  
   LookUpTable = new tSet<int>[NumOfRows];
   Column *c;
   int i,j ;

   for(i=0;i<NumOfColumns;i++)
      if( myColumns.getElementAtSafe(i,c))
      {
	//cout << " process " << i << " c: " << *c << endl;
        for(j=0;j<NumOfRows;j++) if(c->test(j)) LookUpTable[j].addElement(i);
      }

   //printLookUpTable(cout);

   /** keep them sorted in their cost order */

   //cout << " sortAccordingToCost " << endl;

   for(i=0;i<NumOfRows;i++) 
       if( LookUpTable[i].getSize()>0) 
	  sortAccordingToCost(i); //sortAccordingToCost(&LookUpTable[i]);
	  //sortAccordingToCost(j);

   //cout << " end sortAccordingToCost " << endl;

}// end of buildLookUpTable()

/*****************************************************************************/
/** Prints the LookUpTable.  */
/*****************************************************************************/

void SetCoverProblem::printLookUpTable(ostream& os)
{  
   for(int j=0;j<NumOfRows;j++) 
       os << " L[ " << j << "] : " << LookUpTable[j] << endl;

}// end of printLookUpTable()

/*****************************************************************************/
/** Returns the set of indices of columns which covers the given row.  */
/*****************************************************************************/

void SetCoverProblem::getColumnSetCoveringRow(int indx, tSet<int>* ret)
{  
   if(LookUpTable==NULL) cout << " LookUpTable is NULL " << endl;
   if( indx>=0 && indx<NumOfRows ) *ret = LookUpTable[indx];

}// end of getColumnSetCoveringRow()

/*****************************************************************************/
/** Returns the average number of columns per row  */
/*****************************************************************************/

float SetCoverProblem::getAverageNumOfColumns()
{  
   float ret=0;
   for(int i=0;i<NumOfRows;i++) ret += (1.0*LookUpTable[i].getSize()/NumOfRows);

   return ret;

}// end of getAverageNumOfColumns()

/*****************************************************************************/
/** Returns the average number rows per columns */
/*****************************************************************************/

float SetCoverProblem::getAverageNumOfRows()
{  
   float ret=0;
   Column *f;

   for(int i=0;i<NumOfColumns;i++) 
     if( myColumns.getElementAtSafe(i,f) ) ret += (1.0*f->count()/NumOfColumns);

   return ret;

}// end of getAverageNumOfRows()

/*****************************************************************************/
/** Returns the ratio of number of rows over number of columns */
/*****************************************************************************/

float SetCoverProblem::getDensity()
{  
   float ret=0;
   Column *f;

   for(int i=0;i<NumOfColumns;i++) 
     if( myColumns.getElementAtSafe(i,f) ) ret += (1.0*f->count());

   return (ret/(NumOfRows*NumOfColumns));

}// end of getDensity()

